#  testscenarios: extensions to python unittest to allow declarative
#  dependency injection ('scenarios') by tests.
#
# Copyright (c) 2009, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import unittest

import testtools
from testtools.matchers import EndsWith

import testscenarios


class LoggingResult(testtools.TestResult):
    """TestResult that logs its event to a list."""

    def __init__(self, log):
        self._events = log
        super().__init__()

    def startTest(self, test):
        self._events.append(("startTest", test))
        super().startTest(test)

    def stop(self):
        self._events.append("stop")
        super().stop()

    def stopTest(self, test):
        self._events.append(("stopTest", test))
        super().stopTest(test)

    def addFailure(self, test, err=None, details=None):
        self._events.append(("addFailure", test, err))
        super().addFailure(test, err, details)

    def addError(self, test, err=None, details=None):
        self._events.append(("addError", test, err))
        super().addError(test, err, details)

    def addSkip(self, test, reason=None, details=None):
        # Extract reason from details if not provided directly
        if reason is None and details and "reason" in details:
            reason = details["reason"].as_text()
        self._events.append(("addSkip", test, reason))
        super().addSkip(test, reason, details)

    def addSuccess(self, test, details=None):
        self._events.append(("addSuccess", test))
        super().addSuccess(test, details)

    def startTestRun(self):
        self._events.append("startTestRun")
        super().startTestRun()

    def stopTestRun(self):
        self._events.append("stopTestRun")
        super().stopTestRun()

    def done(self):
        self._events.append("done")
        super().done()

    def tags(self, new_tags, gone_tags):
        self._events.append(("tags", new_tags, gone_tags))
        super().tags(new_tags, gone_tags)

    def time(self, a_datetime):
        self._events.append(("time", a_datetime))
        super().time(a_datetime)


class TestTestWithScenarios(testtools.TestCase):
    scenarios = testscenarios.scenarios.per_module_scenarios(
        "impl", (("unittest", "unittest"), ("unittest2", "unittest2"))
    )

    @property
    def Implementation(self):
        if isinstance(self.impl, tuple):
            self.skipTest("import failed - module not installed?")

        class Implementation(testscenarios.WithScenarios, self.impl.TestCase):
            pass

        return Implementation

    def test_no_scenarios_no_error(self):
        class ReferenceTest(self.Implementation):
            def test_pass(self):
                pass

        test = ReferenceTest("test_pass")
        result = unittest.TestResult()
        test.run(result)
        self.assertTrue(result.wasSuccessful())
        self.assertEqual(1, result.testsRun)

    def test_with_one_scenario_one_run(self):
        class ReferenceTest(self.Implementation):
            scenarios = [("demo", {})]

            def test_pass(self):
                pass

        test = ReferenceTest("test_pass")
        log = []
        result = LoggingResult(log)
        test.run(result)
        self.assertTrue(result.wasSuccessful())
        self.assertEqual(1, result.testsRun)
        self.expectThat(log[0][1].id(), EndsWith("ReferenceTest.test_pass(demo)"))

    def test_with_two_scenarios_two_run(self):
        class ReferenceTest(self.Implementation):
            scenarios = [("1", {}), ("2", {})]

            def test_pass(self):
                pass

        test = ReferenceTest("test_pass")
        log = []
        result = LoggingResult(log)
        test.run(result)
        self.assertTrue(result.wasSuccessful())
        self.assertEqual(2, result.testsRun)
        self.expectThat(log[0][1].id(), EndsWith("ReferenceTest.test_pass(1)"))
        self.expectThat(log[4][1].id(), EndsWith("ReferenceTest.test_pass(2)"))

    def test_attributes_set(self):
        class ReferenceTest(self.Implementation):
            scenarios = [("1", {"foo": 1, "bar": 2}), ("2", {"foo": 2, "bar": 4})]

            def test_check_foo(self):
                self.assertEqual(self.foo * 2, self.bar)

        test = ReferenceTest("test_check_foo")
        log = []
        result = LoggingResult(log)
        test.run(result)
        self.assertTrue(result.wasSuccessful())
        self.assertEqual(2, result.testsRun)

    def test_scenarios_attribute_cleared(self):
        class ReferenceTest(self.Implementation):
            scenarios = [("1", {"foo": 1, "bar": 2}), ("2", {"foo": 2, "bar": 4})]

            def test_check_foo(self):
                self.assertEqual(self.foo * 2, self.bar)

        test = ReferenceTest("test_check_foo")
        log = []
        result = LoggingResult(log)
        test.run(result)
        self.assertTrue(result.wasSuccessful())
        self.assertEqual(2, result.testsRun)
        self.assertNotEqual(None, test.scenarios)
        self.assertEqual(None, log[0][1].scenarios)
        self.assertEqual(None, log[4][1].scenarios)

    def test_countTestCases_no_scenarios(self):
        class ReferenceTest(self.Implementation):
            def test_check_foo(self):
                pass

        test = ReferenceTest("test_check_foo")
        self.assertEqual(1, test.countTestCases())

    def test_countTestCases_empty_scenarios(self):
        class ReferenceTest(self.Implementation):
            scenarios = []

            def test_check_foo(self):
                pass

        test = ReferenceTest("test_check_foo")
        self.assertEqual(1, test.countTestCases())

    def test_countTestCases_1_scenarios(self):
        class ReferenceTest(self.Implementation):
            scenarios = [("1", {"foo": 1, "bar": 2})]

            def test_check_foo(self):
                pass

        test = ReferenceTest("test_check_foo")
        self.assertEqual(1, test.countTestCases())

    def test_countTestCases_2_scenarios(self):
        class ReferenceTest(self.Implementation):
            scenarios = [("1", {"foo": 1, "bar": 2}), ("2", {"foo": 2, "bar": 4})]

            def test_check_foo(self):
                pass

        test = ReferenceTest("test_check_foo")
        self.assertEqual(2, test.countTestCases())

    def test_debug_2_scenarios(self):
        log = []

        class ReferenceTest(self.Implementation):
            scenarios = [("1", {"foo": 1, "bar": 2}), ("2", {"foo": 2, "bar": 4})]

            def test_check_foo(self):
                log.append(self)

        test = ReferenceTest("test_check_foo")
        test.debug()
        self.assertEqual(2, len(log))
        self.assertEqual(None, log[0].scenarios)
        self.assertEqual(None, log[1].scenarios)
        self.assertNotEqual(log[0].id(), log[1].id())



#  testscenarios: extensions to python unittest to allow declarative
#  dependency injection ('scenarios') by tests.
# Copyright (c) 2009, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

"""Example TestScenario."""

from testscenarios import TestWithScenarios


scenario1 = ("basic", {"attribute": "value"})
scenario2 = ("advanced", {"attribute": "value2"})


class SampleWithScenarios(TestWithScenarios):
    scenarios = [scenario1, scenario2]

    def test_demo(self):
        self.assertIsInstance(self.attribute, str)



#  testscenarios: extensions to python unittest to allow declarative
#  dependency injection ('scenarios') by tests.
#
# Copyright (c) 2009, Robert Collins <robertc@robertcollins.net>
# Copyright (c) 2010, 2011 Martin Pool <mbp@sourcefrog.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "apply_scenario",
    "apply_scenarios",
    "generate_scenarios",
    "load_tests_apply_scenarios",
    "multiply_scenarios",
]

from itertools import (
    product,
)
import sys

from testtools.testcase import clone_test_with_new_id
from testtools import iterate_tests


def apply_scenario(scenario, test):
    """Apply scenario to test.

    :param scenario: A tuple (name, parameters) to apply to the test. The test
        is cloned, its id adjusted to have (name) after it, and the parameters
        dict is used to update the new test.
    :param test: The test to apply the scenario to. This test is unaltered.
    :return: A new test cloned from test, with the scenario applied.
    """
    name, parameters = scenario
    scenario_suffix = "(" + name + ")"
    newtest = clone_test_with_new_id(test, test.id() + scenario_suffix)
    test_desc = test.shortDescription()
    if test_desc is not None:
        newtest_desc = "%(test_desc)s %(scenario_suffix)s" % vars()
        newtest.shortDescription = lambda: newtest_desc
    for key, value in parameters.items():
        setattr(newtest, key, value)
    return newtest


def apply_scenarios(scenarios, test):
    """Apply many scenarios to a test.

    :param scenarios: An iterable of scenarios.
    :param test: A test to apply the scenarios to.
    :return: A generator of tests.
    """
    for scenario in scenarios:
        yield apply_scenario(scenario, test)


def generate_scenarios(test_or_suite):
    """Yield the tests in test_or_suite with scenario multiplication done.

    TestCase objects with no scenarios specified are yielded unaltered. Tests
    with scenarios are not yielded at all, instead the results of multiplying
    them by the scenarios they specified gets yielded.

    :param test_or_suite: A TestCase or TestSuite.
    :return: A generator of tests - objects satisfying the TestCase protocol.
    """
    for test in iterate_tests(test_or_suite):
        scenarios = getattr(test, "scenarios", None)
        if scenarios:
            for newtest in apply_scenarios(scenarios, test):
                newtest.scenarios = None
                yield newtest
        else:
            yield test


def load_tests_apply_scenarios(*params):
    """Adapter test runner load hooks to call generate_scenarios.

    If this is referenced by the `load_tests` attribute of a module, then
    testloaders that implement this protocol will automatically arrange for
    the scenarios to be expanded. This can be used instead of using
    TestWithScenarios.

    Two different calling conventions for load_tests have been used, and this
    function should support both. Python 2.7 passes (loader, standard_tests,
    pattern), and bzr used (standard_tests, module, loader).

    :param loader: A TestLoader.
    :param standard_test: The test objects found in this module before
        multiplication.
    """
    if getattr(params[0], "suiteClass", None) is not None:
        loader, standard_tests, pattern = params
    else:
        standard_tests, module, loader = params
    result = loader.suiteClass()
    result.addTests(generate_scenarios(standard_tests))
    return result


def multiply_scenarios(*scenarios):
    """Multiply two or more iterables of scenarios.

    It is safe to pass scenario generators or iterators.

    :returns: A list of compound scenarios: the cross-product of all
        scenarios, with the names concatenated and the parameters
        merged together.
    """
    result = []
    scenario_lists = map(list, scenarios)
    for combination in product(*scenario_lists):
        names, parameters = zip(*combination)
        scenario_name = ",".join(names)
        scenario_parameters = {}
        for parameter in parameters:
            scenario_parameters.update(parameter)
        result.append((scenario_name, scenario_parameters))
    return result


def per_module_scenarios(attribute_name, modules):
    """Generate scenarios for available implementation modules.

    This is typically used when there is a subsystem implemented, for
    example, in both Python and C, and we want to apply the same tests to
    both, but the C module may sometimes not be available.

    Note: if the module can't be loaded, the sys.exc_info() tuple for the
    exception raised during import of the module is used instead of the module
    object. A common idiom is to check in setUp for that and raise a skip or
    error for that case. No special helpers are supplied in testscenarios as
    yet.

    :param attribute_name: A name to be set in the scenario parameter
        dictionary (and thence onto the test instance) pointing to the
        implementation module (or import exception) for this scenario.

    :param modules: An iterable of (short_name, module_name), where
        the short name is something like 'python' to put in the
        scenario name, and the long name is a fully-qualified Python module
        name.
    """
    scenarios = []
    for short_name, module_name in modules:
        try:
            mod = __import__(module_name, {}, {}, [""])
        except BaseException:
            mod = sys.exc_info()
        scenarios.append((short_name, {attribute_name: mod}))
    return scenarios



#  testscenarios: extensions to python unittest to allow declarative
#  dependency injection ('scenarios') by tests.
#
# Copyright (c) 2009, Robert Collins <robertc@robertcollins.net>
# Copyright (c) 2010, 2011 Martin Pool <mbp@sourcefrog.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import unittest

import testtools
from testtools.matchers import EndsWith
from typing import cast


import testscenarios
from testscenarios.scenarios import (
    apply_scenario,
    apply_scenarios,
    generate_scenarios,
    load_tests_apply_scenarios,
    multiply_scenarios,
)


class TestGenerateScenarios(testtools.TestCase):
    def hook_apply_scenarios(self):
        self.addCleanup(
            setattr, testscenarios.scenarios, "apply_scenarios", apply_scenarios
        )
        log = []

        def capture(scenarios, test):
            log.append((scenarios, test))
            return apply_scenarios(scenarios, test)

        testscenarios.scenarios.apply_scenarios = capture
        return log

    def test_generate_scenarios_preserves_normal_test(self):
        class ReferenceTest(unittest.TestCase):
            def test_pass(self):
                pass

        test = ReferenceTest("test_pass")
        log = self.hook_apply_scenarios()
        self.assertEqual([test], list(generate_scenarios(test)))
        self.assertEqual([], log)

    def test_tests_with_scenarios_calls_apply_scenarios(self):
        class ReferenceTest(unittest.TestCase):
            scenarios = [("demo", {})]

            def test_pass(self):
                pass

        test = ReferenceTest("test_pass")
        log = self.hook_apply_scenarios()
        tests = list(generate_scenarios(test))
        self.expectThat(tests[0].id(), EndsWith("ReferenceTest.test_pass(demo)"))
        self.assertEqual([([("demo", {})], test)], log)

    def test_all_scenarios_yielded(self):
        class ReferenceTest(unittest.TestCase):
            scenarios = [("1", {}), ("2", {})]

            def test_pass(self):
                pass

        test = ReferenceTest("test_pass")
        tests = list(generate_scenarios(test))
        self.expectThat(tests[0].id(), EndsWith("ReferenceTest.test_pass(1)"))
        self.expectThat(tests[1].id(), EndsWith("ReferenceTest.test_pass(2)"))

    def test_scenarios_attribute_cleared(self):
        class ReferenceTest(unittest.TestCase):
            scenarios = [("1", {"foo": 1, "bar": 2}), ("2", {"foo": 2, "bar": 4})]

            def test_check_foo(self):
                pass

        test = ReferenceTest("test_check_foo")
        tests = list(generate_scenarios(test))
        for adapted in tests:
            self.assertEqual(None, adapted.scenarios)

    def test_multiple_tests(self):
        class Reference1(unittest.TestCase):
            scenarios = [("1", {}), ("2", {})]

            def test_something(self):
                pass

        class Reference2(unittest.TestCase):
            scenarios = [("3", {}), ("4", {})]

            def test_something(self):
                pass

        suite = unittest.TestSuite()
        suite.addTest(Reference1("test_something"))
        suite.addTest(Reference2("test_something"))
        tests = list(generate_scenarios(suite))
        self.assertEqual(4, len(tests))


class TestApplyScenario(testtools.TestCase):
    def setUp(self):
        super(TestApplyScenario, self).setUp()

        self.scenario_name = "demo"
        self.scenario_attrs = {"foo": "bar"}
        self.scenario = (self.scenario_name, self.scenario_attrs)

        class ReferenceTest(unittest.TestCase):
            def test_pass(self):
                pass

            def test_pass_with_docstring(self):
                """The test that always passes.

                This test case has a PEP 257 conformant docstring,
                with its first line being a brief synopsis and the
                rest of the docstring explaining that this test
                does nothing but pass unconditionally.

                """
                pass

        self.ReferenceTest = ReferenceTest

    def test_sets_specified_id(self):
        raw_test = self.ReferenceTest("test_pass")
        raw_id = "ReferenceTest.test_pass"
        scenario_name = self.scenario_name
        expect_id = f"{raw_id}({scenario_name})"
        modified_test = apply_scenario(self.scenario, raw_test)
        self.expectThat(modified_test.id(), EndsWith(expect_id))

    def test_sets_specified_attributes(self):
        raw_test = self.ReferenceTest("test_pass")
        modified_test = apply_scenario(self.scenario, raw_test)
        self.assertEqual("bar", modified_test.foo)

    def test_appends_scenario_name_to_short_description(self):
        raw_test = self.ReferenceTest("test_pass_with_docstring")
        modified_test = apply_scenario(self.scenario, raw_test)
        raw_doc = cast(str, self.ReferenceTest.test_pass_with_docstring.__doc__)
        raw_desc = raw_doc.split("\n")[0].strip()
        scenario_name = self.scenario_name
        expect_desc = f"{raw_desc} ({scenario_name})"
        self.assertEqual(expect_desc, modified_test.shortDescription())


class TestApplyScenarios(testtools.TestCase):
    def test_calls_apply_scenario(self):
        self.addCleanup(
            setattr, testscenarios.scenarios, "apply_scenario", apply_scenario
        )
        log = []

        def capture(scenario, test):
            log.append((scenario, test))

        testscenarios.scenarios.apply_scenario = capture
        scenarios = ["foo", "bar"]
        list(apply_scenarios(scenarios, "test"))
        self.assertEqual([("foo", "test"), ("bar", "test")], log)

    def test_preserves_scenarios_attribute(self):
        class ReferenceTest(unittest.TestCase):
            scenarios = [("demo", {})]

            def test_pass(self):
                pass

        test = ReferenceTest("test_pass")
        tests = list(apply_scenarios(ReferenceTest.scenarios, test))
        self.assertEqual([("demo", {})], ReferenceTest.scenarios)
        self.assertEqual(ReferenceTest.scenarios, tests[0].scenarios)


class TestLoadTests(testtools.TestCase):
    class SampleTest(unittest.TestCase):
        def test_nothing(self):
            pass

        scenarios = [
            ("a", {}),
            ("b", {}),
        ]

    def test_load_tests_apply_scenarios(self):
        suite = load_tests_apply_scenarios(
            unittest.TestLoader(), [self.SampleTest("test_nothing")], None
        )
        result_tests = list(testtools.iterate_tests(suite))
        self.assertEqual(2, len(result_tests), result_tests)

    def test_load_tests_apply_scenarios_old_style(self):
        """Call load_tests in the way used by bzr."""
        suite = load_tests_apply_scenarios(
            [self.SampleTest("test_nothing")],
            self.__class__.__module__,
            unittest.TestLoader(),
        )
        result_tests = list(testtools.iterate_tests(suite))
        self.assertEqual(2, len(result_tests), result_tests)


class TestMultiplyScenarios(testtools.TestCase):
    def test_multiply_scenarios(self):
        def factory(name):
            for i in "ab":
                yield i, {name: i}

        scenarios = multiply_scenarios(factory("p"), factory("q"))
        self.assertEqual(
            [
                ("a,a", dict(p="a", q="a")),
                ("a,b", dict(p="a", q="b")),
                ("b,a", dict(p="b", q="a")),
                ("b,b", dict(p="b", q="b")),
            ],
            scenarios,
        )

    def test_multiply_many_scenarios(self):
        def factory(name):
            for i in "abc":
                yield i, {name: i}

        scenarios = multiply_scenarios(
            factory("p"), factory("q"), factory("r"), factory("t")
        )
        self.assertEqual(3**4, len(scenarios), scenarios)
        self.assertEqual("a,a,a,a", scenarios[0][0])


class TestPerModuleScenarios(testtools.TestCase):
    def test_per_module_scenarios(self):
        """Generate scenarios for available modules"""
        s = testscenarios.scenarios.per_module_scenarios(
            "the_module",
            [
                ("Python", "testscenarios"),
                ("unittest", "unittest"),
                ("nonexistent", "nonexistent"),
            ],
        )
        self.assertEqual("nonexistent", s[-1][0])
        self.assertIsInstance(s[-1][1]["the_module"], tuple)
        s[-1][1]["the_module"] = None
        self.assertEqual(
            s,
            [
                ("Python", {"the_module": testscenarios}),
                ("unittest", {"the_module": unittest}),
                ("nonexistent", {"the_module": None}),
            ],
        )



#  testscenarios: extensions to python unittest to allow declarative
#  dependency injection ('scenarios') by tests.
#
# Copyright (c) 2009, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import doctest
import unittest

import testscenarios


def test_suite():
    loader = unittest.TestLoader()
    standard_tests = unittest.TestSuite()
    return load_tests(loader, standard_tests, None)


def load_tests(loader, standard_tests, pattern):
    test_modules = [
        "testcase",
        "scenarios",
    ]
    prefix = "testscenarios.tests.test_"
    test_mod_names = [prefix + test_module for test_module in test_modules]
    standard_tests.addTests(loader.loadTestsFromNames(test_mod_names))
    doctest.set_unittest_reportflags(doctest.REPORT_ONLY_FIRST_FAILURE)
    standard_tests.addTest(
        doctest.DocFileSuite("../../README.rst", optionflags=doctest.ELLIPSIS)
    )
    return loader.suiteClass(testscenarios.generate_scenarios(standard_tests))



#  testscenarios: extensions to python unittest to allow declarative
#  dependency injection ('scenarios') by tests.
#
# Copyright (c) 2009, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "TestWithScenarios",
    "WithScenarios",
]

import unittest


from testscenarios.scenarios import generate_scenarios

_doc = """
    When a test object which inherits from WithScenarios is run, and there is a
    non-empty scenarios attribute on the object, the test is multiplied by the
    run method into one test per scenario. For this to work reliably the
    WithScenarios.run method must not be overriden in a subclass (or overridden
    compatibly with WithScenarios).
    """


class WithScenarios(object):
    __doc__ = (
        """A mixin for TestCase with support for declarative scenarios.
    """
        + _doc
    )

    def _get_scenarios(self):
        return getattr(self, "scenarios", None)

    def countTestCases(self):
        scenarios = self._get_scenarios()
        if not scenarios:
            return 1
        else:
            return len(scenarios)

    def debug(self):
        scenarios = self._get_scenarios()
        if scenarios:
            for test in generate_scenarios(self):
                test.debug()
        else:
            return super(WithScenarios, self).debug()

    def run(self, result=None):
        scenarios = self._get_scenarios()
        if scenarios:
            for test in generate_scenarios(self):
                test.run(result)
            return
        else:
            return super(WithScenarios, self).run(result)


class TestWithScenarios(WithScenarios, unittest.TestCase):
    __doc__ = (
        """Unittest TestCase with support for declarative scenarios.
    """
        + _doc
    )



#  testscenarios: extensions to python unittest to allow declarative
#  dependency injection ('scenarios') by tests.
#
# Copyright (c) 2009, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.


"""Support for running tests with different scenarios declaratively

Testscenarios provides clean dependency injection for python unittest style
tests. This can be used for interface testing (testing many implementations via
a single test suite) or for classic dependency injection (provide tests with
dependencies externally to the test code itself, allowing easy testing in
different situations).

See the README for a manual, and the docstrings on individual functions and
methods for details.
"""

__all__ = [
    "TestWithScenarios",
    "WithScenarios",
    "apply_scenario",
    "apply_scenarios",
    "generate_scenarios",
    "load_tests_apply_scenarios",
    "multiply_scenarios",
    "per_module_scenarios",
    "__version__",
]


from testscenarios.scenarios import (  # noqa: E402
    apply_scenario,
    apply_scenarios,
    generate_scenarios,
    load_tests_apply_scenarios,
    multiply_scenarios,
    per_module_scenarios,
)
from testscenarios.testcase import TestWithScenarios, WithScenarios  # noqa: E402


def test_suite():
    import testscenarios.tests  # noqa: F401

    return testscenarios.tests.test_suite()


def load_tests(loader, standard_tests, pattern):
    standard_tests.addTests(loader.loadTestsFromNames(["testscenarios.tests"]))
    return standard_tests


def __get_git_version():
    import os
    import subprocess

    cwd = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir)

    try:
        out = subprocess.check_output(
            ["git", "describe"], stderr=subprocess.STDOUT, cwd=cwd
        )
    except (OSError, subprocess.CalledProcessError):
        return None

    try:
        version = out.strip().decode("utf-8")
    except UnicodeDecodeError:
        return None

    if "-" in version:  # after tag
        # convert version-N-githash to version.postN+githash
        return version.replace("-", ".post", 1).replace("-g", "+git", 1)
    else:
        return version


# same format as sys.version_info: "A tuple containing the five components of
# the version number: major, minor, micro, releaselevel, and serial. All
# values except releaselevel are integers; the release level is 'alpha',
# 'beta', 'candidate', or 'final'. The version_info value corresponding to the
# Python version 2.0 is (2, 0, 0, 'final', 0)."  Additionally we use a
# releaselevel of 'dev' for unreleased under-development code.
#
# If the releaselevel is 'alpha' then the major/minor/micro components are not
# established at this point, and setup.py will use a version of next-$(revno).
# If the releaselevel is 'final', then the tarball will be major.minor.micro.
# Otherwise it is major.minor.micro~$(revno).

try:
    from ._version import __version__, version
except ModuleNotFoundError:
    # package is not installed
    if version := __get_git_version():
        # we're in a git repo
        __version__ = tuple([int(v) if v.isdigit() else v for v in version.split(".")])
    else:
        # we're working with a tarball or similar
        version = "0.0.0"
        __version__ = (0, 0, 0)
